/**
 * fylgja-cli.js
 * 
 * Interactive CLI interface
 */
const { parseCommand } = require('./lang/command_parser');
const logger = require('./utils/logger');
const sigmaSearchHandler = require('./handlers/sigma/sigma_search_handler');
const sigmaDetailsHandler = require('./handlers/sigma/sigma_details_handler');
const sigmaStatsHandler = require('./handlers/sigma/sigma_stats_handler');
const sigmaCreateHandler = require('./handlers/sigma/sigma_create_handler');
const { handleCommand: handleAlerts } = require('./handlers/alerts/alerts_handler');
const { handleCommand: handleCase } = require('./handlers/case/case_handler');
const { handleCommand: handleConfig } = require('./handlers/config/config_handler');
const { handleCommand: handleStats } = require('./handlers/stats/stats_handler');
const readline = require('readline');


const colors = {
  // ANSI color codes
  reset: '\x1b[0m',
  bright: '\x1b[1m',
  dim: '\x1b[2m',
  underscore: '\x1b[4m',
  blink: '\x1b[5m',
  reverse: '\x1b[7m',
  hidden: '\x1b[8m',

  // Foreground colors
  black: '\x1b[30m',
  red: '\x1b[31m',
  green: '\x1b[32m',
  yellow: '\x1b[33m',
  blue: '\x1b[34m',
  magenta: '\x1b[35m',
  cyan: '\x1b[36m',
  white: '\x1b[37m',

  // Background colors
  bgBlack: '\x1b[40m',
  bgRed: '\x1b[41m',
  bgGreen: '\x1b[42m',
  bgYellow: '\x1b[43m',
  bgBlue: '\x1b[44m',
  bgMagenta: '\x1b[45m',
  bgCyan: '\x1b[46m',
  bgWhite: '\x1b[47m'
};

// Create simple color functions
const colorize = {
  blue: (text) => `${colors.blue}${text}${colors.reset}`,
  green: (text) => `${colors.green}${text}${colors.reset}`,
  red: (text) => `${colors.red}${text}${colors.reset}`,
  yellow: (text) => `${colors.yellow}${text}${colors.reset}`,
  cyan: (text) => `${colors.cyan}${text}${colors.reset}`,
  white: (text) => `${colors.white}${text}${colors.reset}`,
  dim: (text) => `${colors.dim}${text}${colors.reset}`,
};

const tableColors = {
  header: colorize.cyan,
  key: colorize.blue,
  statsKey: colorize.yellow,
  count: colorize.green,
  dim: colorize.dim
};

// Import CLI formatters
const {
  formatSigmaStats,
  formatSigmaSearchResults,
  formatSigmaDetails
} = require('./utils/cli_formatters');

// Set logger to CLI mode (prevents console output)
logger.setCliMode(true);

// Try to get version, but provide fallback if package.json can't be found
let version = '1.0.0';
try {
  const packageJson = require('../package.json');
  version = packageJson.version;
} catch (e) {
  console.log('Could not load package.json, using default version');
}

const FILE_NAME = 'fylgja-cli.js';

// ASCII art logo for the CLI
const ASCII_LOGO = `
░▒▓████████▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░      ░▒▓██████▓▒░       ░▒▓█▓▒░░▒▓██████▓▒░  
░▒▓█▓▒░      ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░     ░▒▓█▓▒░░▒▓█▓▒░      ░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░ 
░▒▓█▓▒░      ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░     ░▒▓█▓▒░             ░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░ 
░▒▓██████▓▒░  ░▒▓██████▓▒░░▒▓█▓▒░     ░▒▓█▓▒▒▓███▓▒░      ░▒▓█▓▒░▒▓████████▓▒░ 
░▒▓█▓▒░         ░▒▓█▓▒░   ░▒▓█▓▒░     ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░ 
░▒▓█▓▒░         ░▒▓█▓▒░   ░▒▓█▓▒░     ░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░▒▓█▓▒░░▒▓█▓▒░ 
░▒▓█▓▒░         ░▒▓█▓▒░   ░▒▓████████▓▒░▒▓██████▓▒░ ░▒▓██████▓▒░░▒▓█▓▒░░▒▓█▓▒░ 
`;

// Command history array
let commandHistory = [];
let historyIndex = -1;

// Create the readline interface
const rl = readline.createInterface({
  input: process.stdin,
  output: process.stdout,
  completer: completer,
  prompt: 'fylgja> '
});

/**
 * Command auto-completion function
 * @param {string} line Current command line input
 * @returns {Array} Array with possible completions and the substring being completed
 */
function completer(line) {
  const commands = [
    'search sigma',
    'details sigma',
    'sigma stats',
    'stats sigma',
    'search sigma rules where title contains',
    'search sigma where title contains',
    'search sigma rules where tags include',
    'search sigma where tags include',
    'search sigma rules where logsource.category ==',
    'search sigma where logsource.category ==',
    'search sigma rules where modified after',
    'search sigma where modified after',
    'search sigma rules where author is',
    'search sigma where author is',
    'search sigma rules where level is',
    'search sigma where level is',
    'help',
    'exit',
    'quit',
    'clear'
  ];

  const hits = commands.filter((c) => c.startsWith(line));
  return [hits.length ? hits : commands, line];
}

/**
 * Normalize and wrap text for table display
 * @param {string} text Text to normalize and wrap
 * @param {number} maxWidth Maximum width per line
 * @returns {string[]} Array of wrapped lines
 */
function normalizeAndWrap(text, maxWidth) {
  if (!text) return [''];

  // Convert to string and normalize newlines
  text = String(text || '');

  // Replace all literal newlines with spaces
  text = text.replace(/\n/g, ' ');

  // Now apply word wrapping
  if (text.length <= maxWidth) return [text];

  const words = text.split(' ');
  const lines = [];
  let currentLine = '';

  for (const word of words) {
    // Skip empty words (could happen if there were multiple spaces)
    if (!word) continue;

    // If adding this word would exceed max width
    if ((currentLine.length + word.length + (currentLine ? 1 : 0)) > maxWidth) {
      // Push current line if not empty
      if (currentLine) {
        lines.push(currentLine);
        currentLine = '';
      }

      // If the word itself is longer than maxWidth, we need to split it
      if (word.length > maxWidth) {
        let remaining = word;
        while (remaining.length > 0) {
          const chunk = remaining.substring(0, maxWidth);
          lines.push(chunk);
          remaining = remaining.substring(maxWidth);
        }
      } else {
        currentLine = word;
      }
    } else {
      // Add word to current line
      currentLine = currentLine ? `${currentLine} ${word}` : word;
    }
  }

  // Add the last line if not empty
  if (currentLine) {
    lines.push(currentLine);
  }

  return lines;
}


/**
 * Format CLI output
 * @param {Object} data The data to format
 * @param {string} type The type of data (results, details, stats)
 */
function formatOutput(data, type) {
  if (!data) {
    console.log('No data returned from the server.');
    return;
  }

  console.log();

  // word wrapping function
  function normalizeAndWrap(text, maxWidth) {
    if (text === undefined || text === null) return [''];

    // Convert to string and normalize newlines
    const normalized = String(text).replace(/\n/g, ' ');

    // If text fits in one line, return it
    if (normalized.length <= maxWidth) return [normalized];

    const words = normalized.split(' ');
    const lines = [];
    let currentLine = '';

    for (const word of words) {
      // Skip empty words
      if (!word) continue;

      // Check if adding this word would exceed max width
      if ((currentLine.length + word.length + (currentLine ? 1 : 0)) > maxWidth) {
        // Push current line if not empty
        if (currentLine) {
          lines.push(currentLine);
          currentLine = '';
        }

        // Handle words longer than maxWidth by splitting them
        if (word.length > maxWidth) {
          let remaining = word;
          while (remaining.length > 0) {
            const chunk = remaining.substring(0, maxWidth);
            lines.push(chunk);
            remaining = remaining.substring(maxWidth);
          }
        } else {
          currentLine = word;
        }
      } else {
        // Add word to current line
        currentLine = currentLine ? `${currentLine} ${word}` : word;
      }
    }

    // Add the last line if not empty
    if (currentLine) {
      lines.push(currentLine);
    }

    return lines.length ? lines : [''];
  }

  switch (type) {
    case 'search_results':
      // Header
      console.log(
        tableColors.header(
          '#'.padEnd(5) +
          'Title'.padEnd(32) +
          'OS/Product'.padEnd(15) +
          'ID'.padEnd(13)
        )
      );

      if (data.results && data.results.length > 0) {
        data.results.forEach((rule, index) => {
          const num = (index + 1).toString().padEnd(5);
          const title = (rule.title || '').substring(0, 32).padEnd(32);

          // Extract OS/Product from tags or logsource if available
          let osProduct = 'Unknown';
          if (rule.tags && Array.isArray(rule.tags)) {
            // Look for os tags like windows, linux, macos
            const osTags = rule.tags.filter(tag =>
              ['windows', 'linux', 'macos', 'unix', 'azure', 'aws', 'gcp'].includes(tag.toLowerCase())
            );
            if (osTags.length > 0) {
              osProduct = osTags[0].charAt(0).toUpperCase() + osTags[0].slice(1);
            }
          } else if (rule.logsource && rule.logsource.product) {
            osProduct = rule.logsource.product;
          }

          osProduct = osProduct.substring(0, 15).padEnd(15);
          const id = (rule.id || '').substring(0, 13).padEnd(13);

          console.log(`${num}${title}${osProduct}${id}`);
        });
      } else {
        console.log(tableColors.dim('No results found'));
      }

      console.log(tableColors.count(`${data.totalCount || 0} rows in set`));
      break;

    case 'details':
      const detailsKeyWidth = 24;
      const detailsValueWidth = 50;

      for (const [key, value] of Object.entries(data)) {
        if (typeof value !== 'object' || value === null) {
          const formattedKey = tableColors.key(key.padEnd(detailsKeyWidth - 2));

          // Handle wrapping
          const lines = normalizeAndWrap(value, detailsValueWidth);

          // Print first line with the key
          console.log(`${formattedKey} ${lines[0]}`);

          // Print additional lines if there are any
          for (let i = 1; i < lines.length; i++) {
            console.log(`${' '.repeat(detailsKeyWidth)} ${lines[i]}`);
          }
        }
      }
      break;

    case 'stats':
      const statsMetricWidth = 25;

      for (const [key, value] of Object.entries(data)) {
        const formattedKey = tableColors.statsKey(key.padEnd(statsMetricWidth - 2));
        const formattedValue = String(value || '');

        console.log(`${formattedKey} ${formattedValue}`);
      }
      break;

    default:
      console.log(JSON.stringify(data, null, 2));
  }
}

/**
 * Parse out any basic search keywords from a complexSearch query
 * This helps with the search commands that don't quite match the expected format
 * @param {string} input The complex search query
 * @returns {string} Extracted keywords
 */
function extractSearchKeywords(input) {
  if (!input) return '';

  // Try to extract keywords from common patterns
  if (input.includes('title contains')) {
    const match = input.match(/title\s+contains\s+["']([^"']+)["']/i);
    if (match) return match[1];
  }

  if (input.includes('tags include')) {
    const match = input.match(/tags\s+include\s+(\S+)/i);
    if (match) return match[1];
  }

  // Default - just return the input as is
  return input;
}

/**
 * Process a command from the CLI
 * @param {string} input User input command
 */
async function processCommand(input) {
  try {
    // Skip empty commands
    if (!input.trim()) {
      rl.prompt();
      return;
    }

    // Special CLI commands
    if (input.trim().toLowerCase() === 'exit' || input.trim().toLowerCase() === 'quit') {
      console.log('Goodbye!');
      rl.close();
      process.exit(0);
    }

    if (input.trim().toLowerCase() === 'clear') {
      console.clear();
      rl.prompt();
      return;
    }

    // Special case for simple search
    if (input.trim().match(/^search\s+sigma\s+(.+)$/i) && !input.trim().toLowerCase().includes('where') && !input.trim().toLowerCase().includes('with')) {
      const keyword = input.trim().match(/^search\s+sigma\s+(.+)$/i)[1];

      // Add to command history
      commandHistory.push(input);
      historyIndex = commandHistory.length;

      // Create fake command object
      const command = {
        text: keyword,
        user_id: 'cli_user',
        user_name: 'cli_user',
        command: '/fylgja',
        channel_id: 'cli',
        channel_name: 'cli'
      };

      // Create custom respond function
      const respond = createRespondFunction('search', 'sigma', [keyword]);

      console.log(`Executing: module=sigma, action=search, params=[${keyword}]`);

      try {
        await sigmaSearchHandler.handleCommand(command, respond);
      } catch (error) {
        console.error(`Error: ${error.message}`);
        logger.error(`${FILE_NAME}: Command execution error: ${error.message}`);
        logger.debug(`${FILE_NAME}: Error stack: ${error.stack}`);
        rl.prompt();
      }

      return;
    }

    // Special case for complex search
    const complexSearchMatch = input.trim().match(/^search\s+sigma\s+(rules\s+|detections\s+)?(where|with)\s+(.+)$/i);
    if (complexSearchMatch) {
      const complexQuery = complexSearchMatch[3];

      // Add to command history
      commandHistory.push(input);
      historyIndex = commandHistory.length;

      // Create fake command object
      const command = {
        text: complexQuery,
        user_id: 'cli_user',
        user_name: 'cli_user',
        command: '/fylgja',
        channel_id: 'cli',
        channel_name: 'cli'
      };

      // Create custom respond function
      const respond = createRespondFunction('complexSearch', 'sigma', [complexQuery]);

      console.log(`Executing: module=sigma, action=complexSearch, params=[${complexQuery}]`);

      try {
        await sigmaSearchHandler.handleComplexSearch(command, respond);
      } catch (error) {
        console.error(`Error: ${error.message}`);
        logger.error(`${FILE_NAME}: Command execution error: ${error.message}`);
        logger.debug(`${FILE_NAME}: Error stack: ${error.stack}`);
        rl.prompt();
      }

      return;
    }

    // Add to command history
    commandHistory.push(input);
    historyIndex = commandHistory.length;

    // Parse command using existing parser
    const parsedCommand = await parseCommand(input);

    if (!parsedCommand.success) {
      console.log(parsedCommand.message || "Command not recognized. Type 'help' for usage.");
      rl.prompt();
      return;
    }

    // Extract the command details
    const { action, module, params } = parsedCommand.command;

    // Only show execution info to the user, not sending to logger
    console.log(`Executing: module=${module}, action=${action}, params=[${params}]`);

    // Create fake command object similar to Slack's
    const command = {
      text: Array.isArray(params) && params.length > 0 ? params[0] : input,
      user_id: 'cli_user',
      user_name: 'cli_user',
      command: '/fylgja',
      channel_id: 'cli',
      channel_name: 'cli'
    };

    // Special handling for complexSearch to extract keywords
    if (action === 'complexSearch' && module === 'sigma' && params.length > 0) {
      // Try to extract keywords from complex queries
      const searchTerms = extractSearchKeywords(params[0]);
      command.text = searchTerms || params[0];
    }

    // Create custom respond function for CLI
    const respond = createRespondFunction(action, module, params);

    try {
      switch (module) {
        case 'sigma':
          switch (action) {
            case 'search':
              await sigmaSearchHandler.handleCommand(command, respond);
              break;

            case 'complexSearch':
              await sigmaSearchHandler.handleComplexSearch(command, respond);
              break;

            case 'details':
              await sigmaDetailsHandler.handleCommand(command, respond);
              break;

            case 'stats':
              await sigmaStatsHandler.handleCommand(command, respond);
              break;

            case 'create':
              await sigmaCreateHandler.handleCommand(command, respond);
              break;

            default:
              console.log(`Unknown Sigma action: ${action}`);
              rl.prompt();
          }
          break;

        case 'alerts':
          await handleAlerts(command, respond);
          break;

        case 'case':
          await handleCase(command, respond);
          break;

        case 'config':
          await handleConfig(command, respond);
          break;

        case 'stats':
          await handleStats(command, respond);
          break;

        case 'help':
          displayHelp();
          rl.prompt();
          break;

        default:
          console.log(`Unknown module: ${module}`);
          rl.prompt();
      }
    } catch (error) {
      console.error(`Error: ${error.message}`);
      // Log to file but not console
      logger.error(`${FILE_NAME}: Command execution error: ${error.message}`);
      logger.debug(`${FILE_NAME}: Error stack: ${error.stack}`);
      rl.prompt();
    }
  } catch (error) {
    console.error(`Fatal error: ${error.message}`);
    // Log to file but not console
    logger.error(`${FILE_NAME}: Fatal error: ${error.message}`);
    logger.debug(`${FILE_NAME}: Error stack: ${error.stack}`);
    rl.prompt();
  }
}

/**
 * Custom respond function for handling results
 * @param {string} action The action being performed
 * @param {string} module The module being used
 * @param {Array} params The parameters for the action
 * @returns {Function} A respond function for handling results
 */
function createRespondFunction(action, module, params) {
  // Keep track of whether we're waiting for results
  let isWaitingForResults = false;

  return async (response) => {
    // Check if this is a progress message
    const isProgressMessage =
      typeof response === 'object' &&
      response.text &&
      !response.responseData &&
      (response.text.includes('moment') ||
        response.text.includes('searching') ||
        response.text.includes('processing'));

    if (isProgressMessage) {
      console.log(response.text);
      isWaitingForResults = true;
      return; // Don't show prompt after progress messages
    }

    if (typeof response === 'string') {
      console.log(response);
      rl.prompt();
      return;
    }

    // First check for the responseData property (directly from service)
    if (response.responseData) {
      // Format the data using the appropriate formatter
      if (module === 'sigma') {
        let formattedData;

        if (action === 'search' || action === 'complexSearch') {
          formattedData = formatSigmaSearchResults(response.responseData);
          formatOutput(formattedData, 'search_results');
        } else if (action === 'details') {
          formattedData = formatSigmaDetails(response.responseData);
          formatOutput(formattedData, 'details');
        } else if (action === 'stats') {
          formattedData = formatSigmaStats(response.responseData);
          formatOutput(formattedData, 'stats');
        } else {
          console.log(JSON.stringify(response.responseData, null, 2));
        }
      } else {
        // For other modules, just display the JSON
        console.log(JSON.stringify(response.responseData, null, 2));
      }
    }
    // Fallback for text-only responses
    else if (response.text) {
      console.log(response.text);
    } else {
      console.log('Command completed successfully.');
    }

    // Reset waiting state and show prompt after results
    isWaitingForResults = false;
    rl.prompt();
  };
}

/**
 * Display help text
 */
function displayHelp() {
  const helpText = `
Fylgja CLI Help

Basic Sigma Commands:
- search sigma <keyword> - Search for Sigma rules by keyword
- details sigma <rule_id> - Get details about a specific Sigma rule
- stats sigma - Get statistics about Sigma rules database

Advanced Sigma Search Commands:
- search sigma where title contains "ransomware" - Search by title
- search sigma where tags include privilege_escalation - Search by tags
- search sigma where logsource.category == "process_creation" - Search by log source
- search sigma where modified after 2024-01-01 - Search by modification date
- search sigma rules where title contains "ransomware" - Alternative syntax
- search sigma rules where tags include privilege_escalation - Alternative syntax

CLI Commands:
- exit or quit - Exit the CLI
- clear - Clear the terminal screen
- help - Display this help text
  `;

  console.log(helpText);
}
/**
 * Start the CLI application
 */
function startCLI() {
  console.log(ASCII_LOGO);
  console.log(`Fylgja CLI v${version} - Interactive SIEM Management Tool`);
  console.log(`Type 'help' for usage information or 'exit' to quit\n`);

  // Set up key bindings for history navigation
  rl._writeToOutput = function _writeToOutput(stringToWrite) {
    if (stringToWrite === '\\u001b[A' || stringToWrite === '\\u001b[B') {
      // Don't output control characters for up/down arrows
      return;
    }
    rl.output.write(stringToWrite);
  };

  // Set up key listeners for history
  rl.input.on('keypress', (char, key) => {
    if (key && key.name === 'up') {
      if (historyIndex > 0) {
        historyIndex--;
        rl.line = commandHistory[historyIndex];
        rl.cursor = rl.line.length;
        rl._refreshLine();
      }
    } else if (key && key.name === 'down') {
      if (historyIndex < commandHistory.length - 1) {
        historyIndex++;
        rl.line = commandHistory[historyIndex];
        rl.cursor = rl.line.length;
        rl._refreshLine();
      } else if (historyIndex === commandHistory.length - 1) {
        historyIndex = commandHistory.length;
        rl.line = '';
        rl.cursor = 0;
        rl._refreshLine();
      }
    }
  });

  rl.prompt();

  rl.on('line', async (line) => {
    await processCommand(line.trim());
  });

  rl.on('close', () => {
    console.log('Goodbye!');
    process.exit(0);
  });
}

// Check if running directly
if (require.main === module) {
  startCLI();
} else {
  // Export functions for integration with main app
  module.exports = {
    startCLI
  };
}